@@ -14,6 +14,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <dirent.h>
+#include <errno.h>
 #include <string.h>
 #include <stdio.h>
 #include <popt-gnome.h>
@@ -104,7 +105,7 @@ static guchar * gfi_get_species_name (const guchar *fullname, const guchar *fami
  */
 
 static gboolean gfi_debug = FALSE;
-static gboolean gfi_recursive = TRUE;
+static gboolean gfi_recursive = FALSE;
 static gboolean gfi_clean = FALSE;
 static gboolean gfi_usermap = TRUE;
 static gboolean gfi_dynamicmap = FALSE;
@@ -168,6 +169,7 @@ int main (int argc, char ** argv)
 	char ** args;
 	FILE *of;
 	gint i;
+        int result;
 
 	/* Initialize dictionaries */
 
@@ -179,8 +181,15 @@ int main (int argc, char ** argv)
 	/* Parse arguments */
 
 	ctx = poptGetContext (NULL, argc, argv, options, 0);
-	g_return_val_if_fail (ctx != NULL, 1);
-	g_return_val_if_fail (poptGetNextOpt (ctx) == -1, 1);
+
+	result = poptGetNextOpt (ctx);
+	if (result != -1) { 
+                fprintf(stderr, "%s: %s: %s\n",
+                        "gnome-print-install",
+                        poptBadOption(ctx, POPT_BADOPTION_NOALIAS),
+                        poptStrerror(result));
+                return 1;
+	}
 	args = (char **) poptGetArgs (ctx);
 
 	/* Step 1: Read existing fontmap */
@@ -421,6 +430,7 @@ gfi_read_afm_file_data (const guchar *name)
 	fi = NULL;
 
 	if (stat (name, &s) < 0) return NULL;
+	if (!gf_afm_check (name)) return NULL;
 
 	f = fopen (name, "r");
 	if (!f) return NULL;
@@ -509,37 +519,69 @@ gfi_test_file_changed (GFFileEntry *f)
  * Scan directory for font files, update goodafm and goodpfb
  */
 
+#define MAX_RECURSION_DEPTH 32
+
 static void
 gfi_scan_path (const guchar *path, gint level)
 {
+	static ino_t inodes[MAX_RECURSION_DEPTH + 1];
 	struct stat s;
 
-	if (level > 32) return;
+	if (level > MAX_RECURSION_DEPTH) return;
 
 	if (gfi_debug) fprintf (stderr, "Scanning path %s (%s)\n", path, gfi_recursive ? "recursive" : "non-recursive");
 
 	if (!stat (path, &s)) {
 		if (S_ISDIR (s.st_mode)) {
-			DIR * dir;
-			struct dirent * dent;
-
-			dir = opendir (path);
-
-			if (dir) {
-				while ((dent = readdir (dir))) {
-					gchar *fn;
-					if (!strcmp (dent->d_name, ".") || !strcmp (dent->d_name, "..")) continue;
-					fn = g_concat_dir_and_file (path, dent->d_name);
-					if (!stat (fn, &s)) {
-						if (S_ISREG (s.st_mode)) {
-							gfi_try_font_file (fn);
-						} else if (S_ISDIR (s.st_mode) && gfi_recursive) {
-							gfi_scan_path (fn, level + 1);
-						}
-					}
-					g_free (fn);
+			static gboolean cracktest = FALSE;
+			static gboolean crackroot = FALSE;
+			static gboolean crackdev = FALSE;
+			static gboolean crackproc = FALSE;
+			static struct stat s_root, s_dev, s_proc;
+			gboolean seemsok;
+			gint i;
+			/* Check for crack-smoking */
+			if (!cracktest) {
+				crackroot = !stat ("/", &s_root);
+				crackdev = !stat ("/dev", &s_dev);
+				crackproc = !stat ("/proc", &s_proc);
+				cracktest = TRUE;
+			}
+			seemsok = TRUE;
+			/* Successful stat, check if already scanned */
+			for (i = 0; i < level; i++) {
+				if (s.st_ino == inodes[i]) {
+					seemsok = FALSE;
+					break;
 				}
-				closedir (dir);
+			}
+			if (crackroot && (level > 0) && (s.st_ino == s_root.st_ino)) seemsok = FALSE;
+			if (crackdev && (s.st_ino == s_dev.st_ino)) seemsok = FALSE;
+			if (crackproc && (s.st_ino == s_proc.st_ino)) seemsok = FALSE;
+			if (seemsok) {
+				DIR * dir;
+				struct dirent * dent;
+				/* Not yet scanned */
+				inodes[level] = s.st_ino;
+				dir = opendir (path);
+				if (dir) {
+					while ((dent = readdir (dir))) {
+						gchar *fn;
+						if (!strcmp (dent->d_name, ".") || !strcmp (dent->d_name, "..")) continue;
+						fn = g_concat_dir_and_file (path, dent->d_name);
+						if (!stat (fn, &s)) {
+							if (S_ISREG (s.st_mode)) {
+								gfi_try_font_file (fn);
+							} else if (S_ISDIR (s.st_mode) && gfi_recursive) {
+								gfi_scan_path (fn, level + 1);
+							}
+						}
+						g_free (fn);
+					}
+					closedir (dir);
+				}
+			} else {
+				if (gfi_debug) fprintf (stderr, "Circular link or weird path: %s\n", path);
 			}
 		} else if (S_ISREG (s.st_mode)) {
 			gfi_try_font_file (path);
@@ -689,14 +731,14 @@ gfi_process_alias (GFIAliasData *a)
 	afm = afmfd->afm_list->data;
 
 	/* fixme: mess with locale */
-	afmversion = atof (afm->version);
+	afmversion = (afm->version) ? atof (afm->version) : 1.0;
 
 	/* Search, whether we have same or better pfb */
 	for (l = afmfd->pfb_list; l != NULL; l = l->next) {
 		GFIFileData *fd;
 		gdouble pfbversion;
 		fd = l->data;
-		pfbversion = atof (fd->version);
+		pfbversion = (fd->version) ? atof (fd->version) : 1.0;
 		/* If we have original pfb with same or higher version, we shouldn't use type1 alias */
 		if (pfbversion >= afmversion) return;
 	}
@@ -715,13 +757,13 @@ gfi_process_alias (GFIAliasData *a)
 	if (l == NULL) return;
 
 	/* fixme: mess with locale */
-	pfbversion = atof (pfb->version);
+	pfbversion = (pfb->version) ? atof (pfb->version) : 1.0;
 
 	for (l = pfbfd->pfb_list; l != NULL; l = l->next) {
 		GFIFileData *fd;
 		gdouble v;
 		fd = l->data;
-		v = atof (fd->version);
+		v = (fd->version) ? atof (fd->version) : 1.0;
 		if (v > pfbversion) {
 			pfbversion = v;
 			pfb = fd;
@@ -808,7 +850,7 @@ gfi_build_font (GFIFontData *fd)
 		GFIFileData *d;
 		gdouble v;
 		d = (GFIFileData *) l->data;
-		v = atof (d->version);
+		v = (d->version) ? atof (d->version) : 1.0;
 		if (v > pfbversion) {
 			pfbversion = v;
 			pfbdata = d;
@@ -826,7 +868,7 @@ gfi_build_font (GFIFontData *fd)
 		GFIFileData *d;
 		gdouble v;
 		d = (GFIFileData *) l->data;
-		v = atof (d->version);
+		v = (d->version) ? atof (d->version) : 1.0;
 		if ((v > afmversion) && (v <= pfbversion)) {
 			afmversion = v;
 			afmdata = d;
@@ -923,14 +965,22 @@ gfi_get_output_stream (void)
 
 	umask (022);
 
-	if (!strcmp (gfi_target, "-")) return stdout;
+	if (!strcmp (gfi_target, "-")) {
+		if (gfi_debug) fprintf (stderr, "Writing fontmap to stdout\n");
+		return stdout;
+	}
 	if (!gfi_ensure_directory (gfi_target)) {
 		g_warning ("Cannot create path for %s", gfi_target);
 		return NULL;
 	}
 	ostream = fopen (gfi_target, "w");
-	if (ostream) return ostream;
-	g_warning ("Cannot create output file: %s", gfi_target);
+	if (ostream) {
+		if (gfi_debug) fprintf (stderr, "Writing fontmap to %s\n", gfi_target);
+		return ostream;
+	}
+
+        fprintf (stderr, "gnome-font-install: Cannot open output file %s: %s\n",
+                 gfi_target, g_strerror (errno));
 	return NULL;
 }
 
