@@ -188,6 +188,24 @@ gp_fontmap_load (void)
 		gp_fontmap_load_dir (map, FONTMAPDIR_STATIC);
 	}
 
+	/* Sanity check */
+	if (map->num_fonts < 24) {
+		/* Less than 24 fonts means, you do not have PS ones */
+		if (!stat (DATADIR "/fonts/fontmap2", &s) && S_ISREG (s.st_mode)) {
+			gp_fontmap_load_file (map, DATADIR "/fonts/fontmap2");
+		}
+	}
+	/* More sanity check */
+	if (map->num_fonts < 24) {
+		gchar *filename;
+		/* Less than 24 fonts means, you do not have PS ones */
+		filename = g_concat_dir_and_file (g_get_home_dir (), ".gnome/fonts/fontmap");
+		if (!stat (filename, &s) && S_ISREG (s.st_mode)) {
+			gp_fontmap_load_file (map, filename);
+		}
+		g_free (filename);
+	}
+
 	/* Sort fonts alphabetically */
 	map->fonts = g_slist_sort (map->fonts, gp_fe_sortname);
 
@@ -286,8 +304,17 @@ gp_fontmap_load_file (GPFontMap *map, const guchar *filename)
 			version = xmlGetProp (root, "version");
 			if (version) {
 				if (!strcmp (version, "2.0")) {
+					xmlChar *test;
+					gboolean try;
 					/* We are even right version */
-					gp_fm_load_fonts_2_0 (map, root);
+					try = TRUE;
+					test = xmlGetProp (root, "test");
+					if (test) {
+						struct stat s;
+						if (stat (test, &s) || !S_ISREG (s.st_mode)) try = FALSE;
+						xmlFree (test);
+					}
+					if (try) gp_fm_load_fonts_2_0 (map, root);
 				}
 				xmlFree (version);
 			}
