@@ -82,9 +82,11 @@ gp_gc_new (void)
 	GPGC * gc;
 	GPCtx * ctx;
 
+	ctx = gp_ctx_new ();
+	g_return_val_if_fail (ctx != NULL, NULL);
+
 	gc = g_new (GPGC, 1);
 	gc->refcount = 1;
-	ctx = gp_ctx_new ();
 	gc->ctx = g_slist_prepend (NULL, ctx);
 
 	return gc;
@@ -125,6 +127,7 @@ gp_gc_reset (GPGC * gc)
 	}
 
 	ctx = gp_ctx_new ();
+	g_return_if_fail (ctx != NULL);
 
 	gc->ctx = g_slist_prepend (NULL, ctx);
 }
@@ -152,10 +155,8 @@ gp_gc_grestore (GPGC * gc)
 	g_return_val_if_fail (gc != NULL, -1);
 	g_return_val_if_fail (gc->ctx->next != NULL, -1);
 
-	tmp = gc->ctx;
-	gp_ctx_destroy (tmp->data);
-	gc->ctx = tmp->next;
-	g_slist_free_1 (tmp);
+	gp_ctx_destroy ((GPCtx *) gc->ctx->data);
+	gc->ctx = g_slist_remove (gc->ctx, gc->ctx->data);
 
 	return 0;
 }
